"""
This module performs feature engineering on StructuredData
"""

import pandas as pd
import math

from epicas import preprocessor, utils, settings

class EpiData:
    def __init__(self, StructuredData, y, df=None, disease='generic', **kwargs):
        """
        # Arguments:

            - StructuredData

            - y = str. This is your target variable. Choose either:
                + 'incidence'
                + 'prevalence'
                + 'recovered'
                + 'death'

            - disease = str. If unspecified, 'generic' is passed. If disease is
            unknown or in doubt, use generic options:
                + 'generic'
                + 'generic_aerosol'
                + 'generic_body_fluid'
                + 'generic_fecal_oral'
                + 'generic_respiratory'
                + 'generic_respiratory_droplet'

            or choose from the list of well-known based on similarities:
                + 'influenza'
                + 'covid19'
                + 'covid19_alpha'
                + 'covid19_delta'
                + 'sars'
                + 'mers'
                + 'common_cold'
                + 'ebola'
                + 'measles'
                + 'mump'
                + 'hiv'
                + 'hantavirus'
                + 'polio'
                + 'chickenpox'
        """

        self.StructuredData = StructuredData

        if isinstance(df, pd.DataFrame):
            self.df = df
        else:
            self.df = self.StructuredData.df

        self._resort(self.df)
        self.target = y
        self.disease = disease

        if 'location' not in self.df.columns.values.tolist():
            self.df.location = 'single'

        self.headers = StructuredData.df.columns.values.tolist()
        self.features = self._list_of_features(self.headers, self.target)
        self.epi_params = settings.static_params()['infectious_info']

        self._validate(self.headers, self.disease, self.epi_params)


    def __str__(self):
        return f'EpiData{list(self.df.columns)}\n\n' + \
                f'{self.df}\n\n'


    def corr(self, **kwargs):
        """ Calculate correlation between explanatory variables and target
        variables
        """

        self._corr(self.df, self.headers, self.target, self.features, **kwargs)


    def stats(self):
        """ Examine EpiData for NaNs """

        utils.count_na(self.df)


    def lag_reduction(self, subset=None, mid_point=0, sliding_window=None,
        verbose=True):
        """ Shift time-series bi-directionally to maximize correlation between
        predictor and target

        # Arguments:

            - subset: list-like or callable. List of explanatory variables
            whose momentums are presumed to precede target variable.

            - mid_point: int. The midpoint of sliding window

            - sliding_window: int or None. Window size (days) to try shifting
            time-series between. If None, the window will be inferred from
            usual incubation periods for specified disease.

            - verbose: bool. set to False to stop reporting shifted orders
        """

        if sliding_window == None:
            sliding_window = \
                self.epi_params[self.disease]['incubation_period'][-1]

        self.df = self._find_optimal_shift(
            subset,
            self.df,
            self.target,
            mid_point,
            sliding_window,
            verbose
        )

        return EpiData(self.StructuredData, self.target, self.df)


    def imputation(self, method='median', **kwargs):
        """ Fill NaN values

        # Arguments:

            - method: str. Choose either:

                + 'median': fill missing values with medians, this option is
                more robust to outliers.

                + 'mean': fill missing values with mean values

                + 'zero': fill missing values with 0

                + 'ffill': fill missing values by propagating last valid
                observations forward to next

                + 'bfill': fill missing values with next available observations
        """
        if method == 'median':
            for var in self.headers:
                if var not in ['date', 'location']:
                    self.df[var] = self.df[var].fillna(self.df[var].median())

        elif method == 'median':
            for var in self.headers:
                if var not in ['date', 'location']:
                    self.df[var] = self.df[var].fillna(self.df[var].mean())

        elif method == 'zero':
            for var in self.headers:
                if var not in ['date', 'location']:
                    self.df[var] = self.df[var].fillna(0)

        elif method == 'ffill':
            for var in self.headers:
                if var not in ['date', 'location']:
                    self.df[var] = self.df[var].fillna(method='ffill')

        elif method == 'bfill':
            for var in self.headers:
                if var not in ['date', 'location']:
                    self.df[var] = self.df[var].fillna(method='bfill')

        else:
            raise ValueError(f'{method} is not a built-in method. See documentation.')

        return EpiData(self.StructuredData, self.target, self.df)


    def feature_selection(self, n, **kwargs):
        # TODO: implement SHAP model explainer and/or LIME

        """ Perform feature selection and retain best n features.

        Take this function with a grain of salt: unless your task is to quickly
        dive into modeling, this is bad as it may not correctly interpret the
        behavior of our model architecture.

        # Arguments:

            - n: int or float. Number of best features that you want to keep.

                + If <1: proportional to total number of features

                + If >=1: number of features to be kept
        """

        max_n = len(self.features)

        if n > max_n or n<=0:
            raise ValueError('Number of features is invalid')

        elif n < 1:
            n = math.ceil(max_n * n)


        correlations = _corr(
            self.df,
            self.headers,
            self.target,
            self.features,
            verbose=False
        )

        top_features = sorted(
            correlations,
            key = correlations.get,
            reverse=True
        )

        top_n_features = top_features[:n]

        for feature in self.features:
            if feature not in top_n_features:
                self.df = self.df.drop(columns = feature)
                if feature in self.StructuredData.variables['static']:
                    self.StructuredData.variables['static'].remove(feature)
                if feature in self.StructuredData.variables['time_series']:
                    self.StructuredData.variables['time_series'].remove(feature)

        return EpiData(self.StructuredData, self.target, self.df)


    def normalization(self, range=[0, 1], subset='all'):
        """ Apply normalization to variables (similar to sklearn.preprocessing
        .MinMaxScaler)

        # Arguments:

            - range: list-like or callable. The range of features to be
            normalized into. Default: 0<=x<=1.

            - subset = str or list-like (callable). The subset of columns
            to be normalized

                + If 'all': all variables will be normalized except target
                variable

                + If 'time-series': all time-series variables will be normalized
                except target variable

                + If 'static': all static variables will be normalized

                + If list-like: list of specific features to be normalized
        """

        excl = ['location', 'date', self.target]
        norm_variables = []

        if subset == 'all':
            norm_variables = self.features

        if subset == 'time_series':
            for var in self.StructuredData.variables['time_series']:
                if var not in excl:
                    norm_variables.append(var)

        if subset == 'static':
            for var in self.StructuredData.variables['static']:
                if var not in excl:
                    norm_variables.append(var)

        elif type(subset) == list:
            norm_variables = subset

        for var in norm_variables:
            x = self.df[var]
            stdev = (x - x.min(axis=0)) / (x.max(axis=0) - x.min(axis=0))
            self.df[var] = stdev * (range[-1] - range[0]) + range[0]

        return EpiData(self.StructuredData, self.target, self.df)


    def outlier_removal(self, variables, upper=None, lower=None, **kwargs):
        """ Clip outliers off EpiData based on set thresholds

        # Arguments:

            - upper: dict or None. Map variable to its upper threshold

            - lower: dict or None. Map variable to its upper threshold
        """
        self._clip(self.df, variables, upper, lower, **kwargs)

        return EpiData(self.StructuredData, self.target, self.df)


    def cumulative_to_incident(self, subset='target', **kwargs):
        """ Convert cumulative records to incident observations

        # Arguments:

            - subset: str or list-like (callable). Choose either:

                + 'target': only convert target variable

                + 'all': convert all variables that are time-series

                + List of specific columns to be converted
        """

        if self.StructuredData.cumulative:

            if subset == 'all':
                for var in self.variables['time_series']:
                    if var not in ['location', 'date']:
                        df[var] = df[var].diff(**kwargs)

            elif subset == 'target':
                df[self.target] = df[self.target].diff(**kwargs)

            else:
                for var in subset:
                    df[var] = df[var].diff(**kwargs)

        return EpiData(self.StructuredData, self.target, self.df)


    def target_to_ma(self, window, **kwargs):
        """ Convert target observations to their moving average (MA) values

        # Arguments:

            - window: int. Fixed window size to calculate rolling average.
        """

        self.df[self.target] = self.df[self.target].rolling(window).mean()

        self.df[self.target] = self.df[self.target].fillna(method='bfill')

        return EpiData(self.StructuredData, self.target, self.df)


    @staticmethod
    def _corr(df, headers, target, features, verbose = True, **kwargs):
        var_tuples = []
        pearson_scores = []
        correlation_table = {}

        for var in features:

                var_tuples.append(f'({var}, {target})')
                score = df[var].corr(df[target], **kwargs)
                pearson_scores.append(score.round(5))

                correlation_table[var] = score

        if verbose:

            for tuple, score in zip(var_tuples, pearson_scores):
                print(f'Correlation of {tuple:40}:{score:>10}')

        return correlation_table


    @staticmethod
    def _validate(headers, disease, params):
        if 'date' not in headers:
            raise ValueError('Date is missing from your data')

        if not any(var in ['incidence', 'prevalence', 'recovered', 'death']
                   for var in headers):

            raise ValueError('Need at least one target value for EpiData')

        if disease not in params:
            raise ValueError(f'Invalid disease argument: "{disease}"')


    @staticmethod
    def _clip(df, upper, lower):
        if upper != None:
            for var in list(upper.keys()):
                df[var].clip(upper=upper[var], inplace=True)

        if lower != None:
            for var in list(lower.keys()):
                df[var].clip(lower=lower[var], inplace=True)


    @staticmethod
    def _list_of_features(headers, target):
        features = []

        for feature in headers:

            if feature not in ['location', 'date', target]:

                features.append(feature)

        return features


    @staticmethod
    def _find_optimal_shift(subset, df, target, mid_point, sliding_window,
                            verbose):
        start = mid_point - int(sliding_window/2)
        end = start + sliding_window
        y = df[target]

        for var in subset:

            best_score = df[var].corr(y)
            best_shift = 0

            for shift in range(start, end):

                shifted = df.groupby('location')[var].shift(shift)
                score = abs(shifted.corr(y))

                if score > best_score:
                    best_score = score
                    best_shift = shift

            if verbose:
                print(f'Optimal shift for {var:<10}:{best_shift:>3}')

            shifted = df.groupby('location')[var].shift(best_shift)
            df = df.drop(columns=[var])
            df[var] = shifted

        return df.dropna(subset=subset).reset_index(drop=True)

    @staticmethod
    def _resort(df):
        return df.sort_values(['location', 'date'], inplace=True)



"""
LSTM, Bidirectional LSTM (Bi-LSTM), and GRU models

Input: Multivariate, multi-step, multiple time-series
-> Output: Multivariate, single-step, multiple time-series
"""

import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras import Input, Model, layers, optimizers
from tensorflow import keras
from tqdm import tqdm
import os
from datetime import date, datetime, timedelta

from .input import InputLayer

os.environ['CUDA_VISIBLE_DEVICES'] = '-1' # comment out to use GPU(s)


class AutoLSTM(InputLayer):
    def __init__(self, EpiData, model=None, **kwargs):
        super().__init__(EpiData)
        self.model = model

        self.n_timesteps = 14

    def fit(self, verbose=True):
        """Fitting the model and automatically search for best hyperparameters
        within expected range

        # Arguments:
            - verbose: bool. Whether function should print relavant details
            about model

        """
        X_train = np.array(
            np.hsplit(self.X_train.to_numpy(), len(self.locations)))
        Y_train = X_train

        X_val = np.array(np.hsplit(self.X_val.to_numpy(), len(self.locations)))
        Y_val = X_val

        X_test = np.array(
            np.hsplit(self.X_test.to_numpy(), len(self.locations)))
        Y_test = X_test

        n_timesteps = self.n_timesteps  # number of observation to look back

        self.model = self._build_LSTM_model(X_train, n_timesteps)

        self.model.compile(
            optimizer=optimizers.Adam(1e-3),
            loss='mae',
            metrics=['accuracy'])

        if verbose:
            self.model.summary()

        for i in tqdm(range(n_timesteps, X_train.shape[1]), desc='LSTM Fitting'):
            X, Y = X_train[:, (i - n_timesteps):i, :], Y_train[:, i:i + 1, :]

            history = self.model.fit(X, Y, shuffle=False, verbose=0)

        for i in range(n_timesteps, X_val.shape[1]):
            X, Y = X_val[:, (i - n_timesteps):i, :], Y_val[:, i:i + 1, :]

            self.model.evaluate(X, Y, verbose=0)

        return AutoLSTM(self.EpiData, self.model)

    def predict(self, target_date):
        """Predict data from fitted model

        # Argument:
            - target_date: str or datetime
        """
        if isinstance(target_date, str):
            target_date = date.fromisoformat(target_date)

        predict_dates = []

        if target_date <= self.dates_test[-1]:

            X = np.array(np.hsplit(self.X_train.to_numpy(), len(self.locations)))

            forecast_range = (target_date - self.dates_train[-1]).days

            for i in range(1, forecast_range + 1):
                predict_dates.append(self.dates_train[-1] + timedelta(days=i))

                Y_hat = self.model.predict(X[:, -self.n_timesteps:, :])

                Y_hat = Y_hat.reshape(Y_hat.shape[0], 1, Y_hat.shape[1])

                X = np.concatenate((X, Y_hat), axis=1)

        elif target_date > self.dates_test[-1]:
            X = np.array(np.hsplit(self.X_test.to_numpy(), len(self.locations)))

            forecast_range = (target_date - self.dates_test[-1]).days

            for i in range(1, forecast_range + 1):
                predict_dates.append(self.dates_test[-1] + timedelta(days=i))

                Y_hat = self.model.predict(X[:, -self.n_timesteps:, :])

                Y_hat = Y_hat.reshape(Y_hat.shape[0], 1, Y_hat.shape[1])

                X = np.concatenate((X, Y_hat), axis=1)

        predict_values = list(X[:, -len(predict_dates):, 0])

        preds = pd.DataFrame(np.array(predict_values).T,
                    columns = self.Y_val.columns.values.tolist())

        preds['date'] = predict_dates

        preds = preds.melt(id_vars=['date'], \
                            var_name='location', \
                            value_name=self.target)\
                    .reindex(['location', 'date', self.target], axis=1)

        preds.date = preds.date.astype('datetime64[D]')

        return preds

    def save(self, path):
        self.model.save(path)

    def load(self, path):
        self.model = tf.keras.models.load_model(path)
        return AutoLSTM(self.EpiData, self.model)

    @staticmethod
    def _build_LSTM_model(X, n_timesteps):
        # hyperparameters
        n_features = X.shape[-1]
        hidden_units = 64

        # functional API
        inputs = Input(shape=(n_timesteps, n_features))
        lstm = layers.LSTM(hidden_units, return_sequences=True)(inputs)
        lstm = layers.LSTM(hidden_units)(lstm)
        model = layers.Dense(n_features)(lstm)

        return Model(inputs, model)


class AutoGRU(InputLayer):
    def __init__(self, EpiData, model=None, **kwargs):
        super().__init__(EpiData)
        self.model = model
        self.n_timesteps = 14

    def fit(self, verbose=True):
        """Fitting the model and automatically search for best hyperparameters
        within expected range

        # Arguments:
            - verbose: bool. Whether function should print relavant details
            about model

        """
        X_train = np.array(
            np.hsplit(self.X_train.to_numpy(), len(self.locations)))
        Y_train = X_train

        X_val = np.array(np.hsplit(self.X_val.to_numpy(), len(self.locations)))
        Y_val = X_val

        X_test = np.array(
            np.hsplit(self.X_test.to_numpy(), len(self.locations)))
        Y_test = X_test

        n_timesteps = self.n_timesteps  # number of observation to look back

        self.model = self._build_GRU_model(X_train, n_timesteps)

        self.model.compile(
            optimizer=optimizers.Adam(1e-3),
            loss='mae',
            metrics=['accuracy'])

        if verbose:
            self.model.summary()

        for i in tqdm(range(n_timesteps, X_train.shape[1]), desc='GRU Fitting'):
            X, Y = X_train[:, (i - n_timesteps):i, :], Y_train[:, i:i + 1, :]

            history = self.model.fit(X, Y, shuffle=False, verbose=0)

        for i in range(n_timesteps, X_val.shape[1]):
            X, Y = X_val[:, (i - n_timesteps):i, :], Y_val[:, i:i + 1, :]

            self.model.evaluate(X, Y, verbose=0)

        return AutoGRU(self.EpiData, self.model)

    def predict(self, target_date):
        """Predict data from fitted model

        # Argument:
            - target_date: str or datetime
        """
        if isinstance(target_date, str):
            target_date = date.fromisoformat(target_date)

        predict_dates = []

        if target_date <= self.dates_test[-1]:

            X = np.array(np.hsplit(self.X_train.to_numpy(), len(self.locations)))

            forecast_range = (target_date - self.dates_train[-1]).days

            for i in range(1, forecast_range + 1):
                predict_dates.append(self.dates_train[-1] + timedelta(days=i))

                Y_hat = self.model.predict(X[:, -self.n_timesteps:, :])

                Y_hat = Y_hat.reshape(Y_hat.shape[0], 1, Y_hat.shape[1])

                X = np.concatenate((X, Y_hat), axis=1)

        else:
            X = np.array(np.hsplit(self.X_test.to_numpy(), len(self.locations)))

            forecast_range = (target_date - self.dates_test[-1]).days

            for i in range(1, forecast_range + 1):
                predict_dates.append(self.dates_test[-1] + timedelta(days=i))

                Y_hat = self.model.predict(X[:, -self.n_timesteps:, :])

                Y_hat = Y_hat.reshape(Y_hat.shape[0], 1, Y_hat.shape[1])

                X = np.concatenate((X, Y_hat), axis=1)

        predict_values = list(X[:, -len(predict_dates):, 0])

        preds = pd.DataFrame(np.array(predict_values).T,
                    columns = self.Y_val.columns.values.tolist())

        preds['date'] = predict_dates

        preds = preds.melt(id_vars=['date'], \
                            var_name='location', \
                            value_name=self.target)\
                    .reindex(['location', 'date', self.target], axis=1)

        preds.date = preds.date.astype('datetime64[D]')

        return preds

    def save(self, path):
        self.model.save(path)

    def load(self, path):
        self.model = tf.keras.models.load_model(path)
        return AutoGRU(self.EpiData, self.model)

    @staticmethod
    def _build_GRU_model(X, n_timesteps):
        # hyperparameters
        n_features = X.shape[-1]
        hidden_units = 64

        # functional API
        inputs = Input(shape=(n_timesteps, n_features))
        gru = layers.GRU(hidden_units, return_sequences=True)(inputs)
        gru = layers.GRU(hidden_units)(gru)
        model = layers.Dense(n_features)(gru)

        return Model(inputs, model)


class AutoBiLSTM(InputLayer):
    def __init__(self, EpiData, model=None, **kwargs):
        super().__init__(EpiData)
        self.model = model
        self.n_timesteps = 14

    def fit(self, verbose=True):
        """Fitting the model and automatically search for best hyperparameters
        within expected range

        # Arguments:
            - verbose: bool. Whether function should print relavant details
            about model

        """
        X_train = np.array(
            np.hsplit(self.X_train.to_numpy(), len(self.locations)))
        Y_train = X_train

        X_val = np.array(np.hsplit(self.X_val.to_numpy(), len(self.locations)))
        Y_val = X_val

        X_test = np.array(
            np.hsplit(self.X_test.to_numpy(), len(self.locations)))
        Y_test = X_test

        n_timesteps = self.n_timesteps  # number of observation to look back

        self.model = self._build_BiLSTM_model(X_train, n_timesteps)

        self.model.compile(
            optimizer=optimizers.Adam(1e-3),
            loss='mae',
            metrics=['accuracy'])

        if verbose:
            self.model.summary()

        for i in tqdm(range(n_timesteps, X_train.shape[1]), desc='Bidirectional LSTM Fitting'):
            X, Y = X_train[:, (i - n_timesteps):i, :], Y_train[:, i:i + 1, :]

            history = self.model.fit(X, Y, shuffle=False, verbose=0)

        for i in range(n_timesteps, X_val.shape[1]):
            X, Y = X_val[:, (i - n_timesteps):i, :], Y_val[:, i:i + 1, :]

            self.model.evaluate(X, Y, verbose=0)

        return AutoBiLSTM(self.EpiData, self.model)

    def predict(self, target_date):
        """Predict data from fitted model

        # Argument:
            - target_date: str or datetime
        """
        if isinstance(target_date, str):
            target_date = date.fromisoformat(target_date)

        predict_dates = []

        if target_date <= self.dates_test[-1]:

            X = np.array(np.hsplit(self.X_train.to_numpy(), len(self.locations)))

            forecast_range = (target_date - self.dates_train[-1]).days

            for i in range(1, forecast_range + 1):
                predict_dates.append(self.dates_train[-1] + timedelta(days=i))

                Y_hat = self.model.predict(X[:, -self.n_timesteps:, :])

                Y_hat = Y_hat.reshape(Y_hat.shape[0], 1, Y_hat.shape[1])

                X = np.concatenate((X, Y_hat), axis=1)

        else:
            X = np.array(np.hsplit(self.X_test.to_numpy(), len(self.locations)))

            forecast_range = (target_date - self.dates_test[-1]).days

            for i in range(1, forecast_range + 1):
                predict_dates.append(self.dates_test[-1] + timedelta(days=i))

                Y_hat = self.model.predict(X[:, -self.n_timesteps:, :])

                Y_hat = Y_hat.reshape(Y_hat.shape[0], 1, Y_hat.shape[1])

                X = np.concatenate((X, Y_hat), axis=1)

        predict_values = list(X[:, -len(predict_dates):, 0])

        preds = pd.DataFrame(np.array(predict_values).T,
                    columns = self.Y_val.columns.values.tolist())

        preds['date'] = predict_dates

        preds = preds.melt(id_vars=['date'], \
                            var_name='location', \
                            value_name=self.target)\
                    .reindex(['location', 'date', self.target], axis=1)

        preds.date = preds.date.astype('datetime64[D]')

        return preds

    def save(self, path):
        self.model.save(path)

    def load(self, path):
        self.model = tf.keras.models.load_model(path)
        return AutoBiLSTM(self.EpiData, self.model)

    @staticmethod
    def _build_BiLSTM_model(X, n_timesteps):
        # hyperparameters
        n_features = X.shape[-1]
        hidden_units = 64

        # functional API
        inputs = Input(shape=(n_timesteps, n_features))

        bi_lstm = layers.Bidirectional(
            layers.LSTM(hidden_units, return_sequences=True))(inputs)

        bi_lstm = layers.Bidirectional(layers.LSTM(hidden_units))(bi_lstm)

        model = layers.Dense(n_features)(bi_lstm)

        return Model(inputs, model)


def _plausible_timesteps(epi_window, max_timestep):
    """Return a list of possible numbers of timesteps to tune the model

    # Arguments:
    - epi_window: tuple or list. Contains minimum incubation period and
    maximum

    - max_timestep: the largest possible number of timesteps that data
    allows
    """

    if max_timestep > epi_window[-1]:
        return list(range(epi_window[0], max_timestep + 1)) + [max_timestep]

    else:
        return list(range(epi_window[0], max_timestep))



from .preprocessor import StructuredData, EpiData
from .blocks import InputLayer, AutoARIMA, AutoLSTM, AutoGRU, AutoBiLSTM, AutoAttention
from .core import Ensemble
from .settings import static_params

__version__ = '0.1.0'



"""
Augmented Dickey–Fuller (ADF) testing for stationarity and ARIMA/SARIMA modeling
"""
from statsmodels.tsa.arima.model import ARIMA, ARIMAResults
from statsmodels.tsa.stattools import adfuller
import numpy as np
import pandas as pd
from datetime import date, datetime, timedelta
from tqdm import tqdm
import warnings
import os

from .input import InputLayer


class AutoARIMA(InputLayer):
    """
    ARIMA = AutoRegression Intergrated Moving Average

    - If training time and/or memory usage is a major concern, try VAR (Vector
    AutoRegression)
    """
    def __init__(self, EpiData, models=None, **kwargs):
        """
        # Arguments:
        - EpiData
        """

        super().__init__(EpiData, **kwargs)

        self.models = models

    def fit(self, verbose=True, **kwargs):
        """Fit train data. The procedure is given below:

        1. Find the maximum differencing order (call this tolerance)

        2. Find the best parameter d by incrementally differencing until
        time-series is stationary (based on Augmented Dickey-Fuller test)

        3. Find the best parameter p (for AR) by running stepwise tests on the
        first location

        4. Pass these two parameters to fit other locations (batches)- if there
        exist. Stepwise testing can be extremely costly if performed on all
        locations and large amount of oberservations. For the time-efficiency,
        we only run it once.

        # Argument:
            - verbose: bool. Set it to True to see ARIMA Results after fitting.
            False by default.
        """

        # TODO: multiprocessing and/or caching

        self.models = []


        Y = np.hsplit(self.Y_train.to_numpy(), len(self.locations))


        for location in tqdm(range(len(Y)), desc='ARIMA Fitting'):

            y = Y[location].flatten()

            with warnings.catch_warnings():

                warnings.filterwarnings('ignore')

                if location == 0:
                    d = self._find_diff_order(y, self.max_timestep)

                    model, p = self._stepwise_AR_variation(
                        y,
                        d,
                        self.max_timestep
                    )

                else:
                    model = ARIMA(y, order=(p, d, 0))\
                        .fit(low_memory=True, **kwargs)

            self.models.append(model)

        if verbose:
            print(self.models[0].summary())

        return AutoARIMA(self.EpiData, self.models)

    def predict(self, target_date, **kwargs):
        """Predict data from fitted models

        # Argument:
            - target_date: str or datetime
        """
        # TODO: multiprocessing

        predict_dates = []
        predict_values = []

        if isinstance(target_date, str):
            target_date = date.fromisoformat(target_date)

        dates = self.dates_train

        last_date_trained_on = dates[-1]

        steps = (target_date - last_date_trained_on).days

        start = len(dates)

        end = start + steps - 1

        predict_dates = [last_date_trained_on + timedelta(days=1 + x) for
                              x in range(steps)]

        for location in tqdm(range(len(self.models)), desc='ARIMA Predicting'):

            forecast = self.models[location].predict(start, end)

            predict_values.append(forecast)

        preds = pd.DataFrame(np.array(predict_values).T,
                            columns = self.Y_train.columns.values.tolist())

        preds['date'] = predict_dates

        preds = preds.melt(id_vars=['date'], \
                            var_name='location', \
                            value_name=self.target)\
                    .reindex(['location', 'date', self.target], axis=1)

        preds.date = preds.date.astype('datetime64[D]')

        return preds

    def save(self, path, **kwargs):
        """Export fitted models to specific path for later use

        # Argument:
            - path: str. the name of the directory to export ARIMAResults
            instances into.
        """
        if not os.path.exists(path):
            os.mkdir(path)

        for i in tqdm(range(len(self.models)), desc='Saving models...'):
            self.models[i].save(path + f'/{self.locations[i]}.pkl')

    def load(self, path):
        fs = os.listdir(path + '/')
        models = []

        for f in fs:
            models.append(ARIMAResults.load(path + '/' + f))

        return AutoARIMA(self.EpiData, models)

    @staticmethod
    def _adf(y):
        """
        Implement Augmented Dickey–Fuller to test for stationarity of series

        # Output:
            - If True, series may be stationary. Otherwise, it may be
            non-stationary

        # References:
        https://www.statsmodels.org/stable/generated/statsmodels.tsa.stattools
        .adfuller.html

        MacKinnon, J.G. 1994. “Approximate asymptotic distribution functions for
        unit-root and cointegration tests. Journal of Business and Economic
        Statistics 12, 167-76.
        """
        adf = adfuller(y)

        if all(adf[0] > val for val in list(adf[4].values())):
            return False  # series may be non-stationary

        else:
            return True  # series may be stationary

    @staticmethod
    def _find_diff_order(y, max_order):
        for d in range(0, max_order + 1):
            if AutoARIMA._adf(np.diff(y, d)):
                return d

    @staticmethod
    def _stepwise_AR_variation(y, d, p_end, p_start=0):
        # TODO: EarlyStopping mechanism
        """Choosing the optimal parameter using AIC and BIC metrics

        # Arguments:
            - y: array. the series

            - dates: array of datetime

            - d: differencing order (obtained from ADF)

            - p_start: int. First p to try with.

            - p_end: int. maximum p

            - early_stopping_delta: float or int. Inspired by EarlyStopping
            callback in Tensorflow, stepwise testing will be stopped if AIC and
            BIC improvements fall under threshold% (i.e., 0.1 for 10%) of the
            previous ones. This helps save time and computational cost, while
            mitigating overfitting.
        """
        optimal_aic = None
        optimal_bic = None
        optimal_model = None
        optimal_p = 0

        for p in range(p_start, p_end + 1):
            model = ARIMA(y, order=(p, d, 0)).fit()

            if (optimal_aic, optimal_bic) == (None, None):
                optimal_aic, optimal_bic, optimal_p = model.aic, model.bic, p
                optimal_model = model

            elif model.aic < optimal_aic and model.bic < optimal_bic:

                optimal_aic, optimal_bic, optimal_p = model.aic, model.bic, p
                optimal_model = model

        return optimal_model, optimal_p



"""
Encode EpiData to input layer
"""

import warnings
import numpy as np
from datetime import datetime


class InputLayer:
    def __init__(self, EpiData, **kwargs):
        # TODO: add arg for custom train-val-test ratio
        """
        # Arguments:

            - EpiData: EpiData obj. The feature engineered data

            - StructuredData: StructuredData obj. Original data

            - target = int. Target variable

            - target_date = int, str, or datetime.

                + If int: the number of days into the future that is relative
                to the last observation.

                + If str: isoformat string to be parsed as date (for example,
                '2021-10-02' is a valid string for Oct 2, 2021)

                + If datetime: this date will be read directly without parsed

            - train_test_ratio = None or float.

                + If float: (between 0 and 1) this will be the ratio of the
                training data proportionally to the EpiData. The rest will be
                splited 50/50 into validation and testing

                + If None: the ratio will be automatically scaled based on
                the size of the EpiData

        """
        self.EpiData = EpiData
        self.StructuredData = self.EpiData.StructuredData
        self.target = self.EpiData.target
        self.disease = self.EpiData.disease
        self.windows = self.EpiData.epi_params[self.disease]['incubation_period']
        self.features = self.EpiData.features
        self.df = self._multi_indexing(self.EpiData.df)

        self.time_series_vars = self._update_variable_list(
            self.EpiData.StructuredData.variables['time_series'],
            self.df,
            self.target
        )

        self.static_vars = self._update_variable_list(
            self.EpiData.StructuredData.variables['static'],
            self.df,
            self.target
        )

        self.X, self.X_train, self.X_val, self.X_test, self.Y,\
            self.Y_train, self.Y_val, self.Y_test, self.dates_train,\
            self.dates_val, self.dates_test, self.max_timestep,\
            self.locations, self.Y_out =\
            self._train_test_split(
                self.df,
                self.target,
                self.windows[-1],
                self.time_series_vars,
                self.static_vars
            )

    def __str__(self):
        return f'EpiData{list(self.df.columns)}\n\n' + \
            f'{self.df}\n\n'

    @staticmethod
    def _infer_ratio(len, max_window):
        """Return split sizes based on length of available data"""

        if len >= 100000:  # training: 99%, validation: 0.5%, testing: 0.5%
            train_cut = int(0.99 * len)
            validation_cut = int(0.995 * len)

        elif len >= 10000:  # training: 90%, validation: 5%, testing: 5%
            train_cut = int(0.9 * len)
            validation_cut = int(0.95 * len)

        else:  # training: 80%, validation: 10%, testing: 10%
            train_cut = int(0.8 * len)
            validation_cut = int(0.9 * len)

        if (len - validation_cut + 1) < max_window:
            train_cut -= max_window - (len - validation_cut + 1)
            validation_cut -= max_window - (len - validation_cut + 1)

        if (validation_cut - train_cut + 1) < max_window:
            compensation = max_window - (validation_cut - train_cut + 1)
            train_cut -= compensation
            validation_cut -= compensation

        if train_cut < max_window or validation_cut < 0:
            return 0, 0
        else:
            return train_cut, validation_cut

    @staticmethod
    def _train_test_split(df, target, max_window, time_series_vars,
                          static_vars):
        """Split dataset into training, validating, and testing NumPy arrays

        # Arguments:
            - df: DataFrame fetched from EpiData

            - target: name of target variable

            - min_window: the minimum number of days to look back

            - max_window: the maximum number of days to look back

            - time_series_vars: list of time_series_variables (except 'date',
            'location' and target variable)

            - static_vars: list of time_series_variables (except 'location')

        # Outputs:
            - train_data: list of (dates, target values, time-series regressors,
            static regressors) to be used for training, one for each location

            - val_data: list of (dates, target values, time-series regressors,
            static regressors) to be used for validation, one for each location

            - test_data: list of (dates, target values, time-series regressors,
            static regressors) to be used for testing, one for each location

        """

        locations = list(set(df.index.values.tolist()))

        X = df.pivot_table(
                index='date',
                columns='location',
                values=[target] + time_series_vars
            ).ffill().bfill()

        dates = X.index.values.astype('datetime64[D]').astype(datetime)
        length = len(dates)

        X = X\
            .reset_index(drop=True)\
            .swaplevel(axis=1)\
            .sort_index(axis=1)\
            .reindex([target] + time_series_vars, axis=1, level=1)

        X_train = X[:int(length/1.25)]
        dates_train = dates[:int(length/1.25)]

        X_val = X[int(length/1.25):int(length/(10/9))]
        dates_val = dates[int(length/1.25):int(length/(10/9))]

        X_test = X[int(length/(10/9)):]
        dates_test = dates[int(length/(10/9)):]

        Y_out = df.pivot_table(
                index='date',
                columns='location',
                values=target
            ).ffill().bfill()

        Y = Y_out.reset_index(drop=True)

        Y_train = Y[:int(length/1.25)]

        Y_val = Y[int(length/1.25):int(length/(10/9))]

        Y_test = Y[int(length/(10/9)):]

        max_timestep = min(max_window, len(dates_val), len(dates_test))

        return X, X_train, X_val, X_test, Y, Y_train, Y_val, Y_test,\
            dates_train, dates_val, dates_test, max_timestep, locations, Y_out

    @staticmethod
    def _multi_indexing(df):
        return df.set_index('location')

    @staticmethod
    def _find_available_length(df, locations):

        if len(locations) != 1:
            max_count = 0

            for location in locations:
                new_count = df.loc[location].shape[0]

                if new_count > max_count:
                    max_count = new_count

        else:
            max_count = df.loc[location[0]].shape[0]

        return max_count

    @staticmethod
    def _update_variable_list(li, df, target):

        cols = []

        for column in li:
            if column in df.columns.values.tolist() and \
                    column not in [target, 'date', 'location']:

                cols.append(column)

        return cols



import pandas as pd
import json

def get_format(path):
    return path.split('.')[-1].lower()


def count_na(df):
    nulls = df.isnull().sum()

    for column in df.columns:
        print(f'NaN / NaT data in {column:30}:' +
            f'{(nulls/len(df)*100)[column].round(1):>5} %')

    return nulls


def read_long_csv(file_input, labels, start_date, tot_pop, dropna, **kwargs):
    col_map = {x: y for y, x in labels.items()}

    df = pd.read_csv(
        file_input,
        usecols=list(labels.values()),
        **kwargs)

    if dropna == True:
        df = df.dropna()
    elif type(dropna) == list:
        df = df.dropna(subset=dropna)

    df = df.rename(columns=col_map)

    if 'tot_pop' not in list(df.columns.values):
        if tot_pop != None:
            df['tot_pop'] = tot_pop

    return df


def read_wide_csv(file_input, labels, start_date, tot_pop, dropna, dropcols,
                wide_label, **kwargs):
    col_map = {x: y for y, x in labels.items()}

    df = pd.read_csv(file_input, **kwargs)

    if dropna == True:
        df = df.dropna()
    elif type(dropna) == list:
        df = df.dropna(subset=dropna)

    if type(dropcols) == list:
        df = df.drop(columns=dropcols)

    df = df.rename(columns=col_map)

    cols = list(labels.keys())

    if 'location' in labels:
        cols.remove('location')

        df = df.melt(
            id_vars=['location'],
            var_name = 'date',
            value_name = wide_label
        )


    else:
        cols.remove('date')

        df = df.melt(
            var_name = 'date',
            value_name = wide_label
        )

    return df


def variables_merge(static_1, time_series_1, static_2, time_series_2):
    if static_1 != None and static_2 != None:
        static = list(set(static_1 + static_2))

    elif static_1 != None:
        static = static_1

    elif static_2 != None:
        static = static_2

    else:
        static = None

    if time_series_1 != None and time_series_2 != None:
        time_series = list(set(time_series_1 + time_series_2))

    elif time_series_1 != None:
        time_series = time_series_1

    elif time_series_2 != None:
        time_series = time_series_2

    else:
        time_series = None

    return static, time_series



"""
Ensemble forecasts from models.
"""

import numpy as np
import pandas as pd
from datetime import date, datetime

from epicas.blocks.input import InputLayer
from epicas.blocks.arima import AutoARIMA
from epicas.blocks.sequence import AutoLSTM, AutoBiLSTM, AutoGRU
from epicas.blocks.attention import AutoAttention


class Ensemble(InputLayer):
    def __init__(self, EpiData, models, target_date):
        """Automatically train models, validates and gets weights for weighted
        ensemble model

        # Arguments:
            - EpiData: feature engineered models

            - models: list. models to use, choose from:

                + 'ARIMA'

                + 'attention'

                + 'LSTM'

                + 'BiLSTM'

                + 'GRU'

            - target_date: str or datetime
        """
        super().__init__(EpiData)

        self.df = self.Y_out.reset_index(drop=False)\
                            .melt(id_vars=['date'],
                                  var_name='location',
                                  value_name=self.target)\
                            .reindex(['location', 'date', self.target], axis=1)

        self.df = self._unweighted_ensemble(
            models,
            self.df,
            EpiData,
            target_date,
            self.target
        )

    def get_predict(self, in_sample=True):
        """Return a Pandas DataFrame of predicted values

        # Arguments:
            - in_sample: bool. If True: including in-sample grouth truth data
            in the final EpiData. Otherwise, only outputting predicted data

        """
        if in_sample:
            return self._in_sample(self.df, self.target)
        else:
            return self._out_of_sample(self.df, self.target)

    def save_predict(self, path, in_sample=True):
        """Return a Pandas DataFrame of predicted values

        # Arguments:
            -path: str. Export dataset as CSV.

            - in_sample: bool. If True: including in-sample grouth truth data
            in the final EpiData. Otherwise, only outputting predicted data

        """
        if in_sample:
            df = self._in_sample(self.df, self.target)
        else:
            df = self._out_of_sample(self.df, self.target)

        df.to_csv(path, index=False)

    def __str__(self):
        return f'{self.df}'

    @staticmethod
    def _run_model(model_name, EpiData, date):

        if model_name.lower() == 'attention':
            output = AutoAttention(EpiData).fit().predict(date)

        elif model_name.lower() == 'arima':
            output = AutoARIMA(EpiData).fit().predict(date)

        elif model_name.lower() == 'lstm':
            output = AutoLSTM(EpiData).fit().predict(date)

        elif model_name.lower() == 'bilstm' or model_name.lower() == 'bi-lstm':
            output = AutoBiLSTM(EpiData).fit().predict(date)

        elif model_name.lower() == 'gru':
            output = AutoGRU(EpiData).fit().predict(date)

        else:
            raise ValueError(f'Invalid model name: {model_name}')

        return output

    @staticmethod
    def _unweighted_ensemble(model_names, df, EpiData, date, target):

        for model_name in model_names:

            output = Ensemble._run_model(model_name, EpiData, date)
            output = output.rename(columns={target: model_name})

            df = df.merge(output, on=['date', 'location'], how='outer')

        df[f'{target}_preds'] = df[model_names].mean(axis=1)

        df = df.drop(columns=model_names)

        return df

    @staticmethod
    def _out_of_sample(df, target):

        df = df[df[target].isnull()]
        df = df.drop(columns=[target])

        df[target] = df[f'{target}_preds']
        
        df = df.drop(columns=[f'{target}_preds'])

        return df

    @staticmethod
    def _in_sample(df, target):

        df[target] = df[target].fillna(df[f'{target}_preds'])
        df = df.drop(columns=[f'{target}_preds'])

        return df



import setuptools

setuptools.setup(
    name='epicas',
    version='0.1.0',
    author='Duy Cao',
    author_email='caominhduy@gmail.com',
    description='An AutoML framework designed for epidemiological forecasting',
    long_description='''
Epicas stands for Epidemiological Forecasting. Epicas is an AutoML framework based on TensorFlow and statsmodels.

For instructions or details about source code, please see at https://github.com/caominhduy/epicas

This package is still under development. Feel free to contact developer at caominhduy@gmail.com if any issue arises.
    ''',
    url='https://github.com/caominhduy/epicas',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
    packages=setuptools.find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=[
        'tensorflow>=2.6.0',
        'keras-self-attention>=0.50.0',
        'pandas>=1.2.4',
        'statsmodels>=0.13.0',
        'tqdm',
        'numpy'
    ],
    keywords=[
        'AutoML',
        'ARIMA',
        'attention',
        'epidemiology',
        'forecast',
        'infectious',
        'machine learning',
        'TensorFlow'
        ],
    python_requires='>=3.7.0'
)



"""
"Attention is all epicas.AutoBiLSTM needs"

Input: Multivariate, multi-step, multiple time-series
-> Output: Multivariate, single-step, multiple time-series
"""

import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras import Input, Model, layers, optimizers
from keras_self_attention import SeqSelfAttention
from tqdm import tqdm
import os
from datetime import date, datetime, timedelta

from .input import InputLayer

class AutoAttention(InputLayer):
    def __init__(self, EpiData, model=None, **kwargs):
        super().__init__(EpiData)
        self.model = model
        self.n_timesteps = 14

    def fit(self, verbose=True):
        """Fitting the model and automatically search for best hyperparameters
        within expected range

        # Arguments:
            - verbose: bool. Whether function should print relavant details
            about model

        """
        X_train = np.array(
            np.hsplit(self.X_train.to_numpy(), len(self.locations)))
        Y_train = X_train

        X_val = np.array(np.hsplit(self.X_val.to_numpy(), len(self.locations)))
        Y_val = X_val

        X_test = np.array(
            np.hsplit(self.X_test.to_numpy(), len(self.locations)))
        Y_test = X_test

        n_timesteps = self.n_timesteps  # number of observation to look back

        self.model = self._build_attention_based_BiLSTM_model(X_train, n_timesteps)

        self.model.compile(
            optimizer=optimizers.Adam(1e-3),
            loss='mae',
            metrics=['accuracy'])

        if verbose:
            self.model.summary()

        for i in tqdm(range(n_timesteps, X_train.shape[1]), desc='Attention-based Bi-LSTM Fitting'):
            X, Y = X_train[:, (i - n_timesteps):i, :], Y_train[:, i:i + 1, :]

            history = self.model.fit(X, Y, shuffle=False, verbose=0)

        for i in range(n_timesteps, X_val.shape[1]):
            X, Y = X_val[:, (i - n_timesteps):i, :], Y_val[:, i:i + 1, :]

            self.model.evaluate(X, Y, verbose=0)

        return AutoAttention(self.EpiData, self.model)

    def predict(self, target_date):
        """Predict data from fitted model

        # Argument:
            - target_date: str or datetime
        """
        if isinstance(target_date, str):
            target_date = date.fromisoformat(target_date)

        predict_dates = []

        if target_date <= self.dates_test[-1]:

            X = np.array(np.hsplit(self.X_train.to_numpy(), len(self.locations)))

            forecast_range = (target_date - self.dates_train[-1]).days

            for i in range(1, forecast_range + 1):
                predict_dates.append(self.dates_train[-1] + timedelta(days=i))

                Y_hat = self.model.predict(X[:, -self.n_timesteps:, :])

                Y_hat = Y_hat.reshape(Y_hat.shape[0], 1, Y_hat.shape[1])

                X = np.concatenate((X, Y_hat), axis=1)

        else:
            X = np.array(np.hsplit(self.X_test.to_numpy(), len(self.locations)))

            forecast_range = (target_date - self.dates_test[-1]).days

            for i in range(1, forecast_range + 1):
                predict_dates.append(self.dates_test[-1] + timedelta(days=i))

                Y_hat = self.model.predict(X[:, -self.n_timesteps:, :])

                Y_hat = Y_hat.reshape(Y_hat.shape[0], 1, Y_hat.shape[1])

                X = np.concatenate((X, Y_hat), axis=1)

        predict_values = list(X[:, -len(predict_dates):, 0])

        preds = pd.DataFrame(np.array(predict_values).T,
                    columns = self.Y_train.columns.values.tolist())

        preds['date'] = predict_dates

        preds = preds.melt(id_vars=['date'], \
                            var_name='location', \
                            value_name=self.target)\
                    .reindex(['location', 'date', self.target], axis=1)

        preds.date = preds.date.astype('datetime64[D]')

        return preds

    def save(self, path):
        self.model.save(path)

    def load(self, path):
        self.model = tf.keras.models.load_model(
            path,
            custom_objects=SeqSelfAttention.get_custom_objects()
        )
        return AutoAttention(self.EpiData, self.model)

    @staticmethod
    def _build_attention_based_BiLSTM_model(X, n_timesteps):
        # hyperparameters
        n_features = X.shape[-1]
        hidden_units = 64

        # functional API
        inputs = Input(shape=(n_timesteps, n_features))

        bi_lstm = layers.Bidirectional(
            layers.LSTM(hidden_units, return_sequences=True))(inputs)

        bi_lstm = layers.Bidirectional(
            layers.LSTM(hidden_units, return_sequences=True))(bi_lstm)

        attention = SeqSelfAttention(
            attention_activation='sigmoid',
            attention_type=SeqSelfAttention.ATTENTION_TYPE_MUL)(bi_lstm)

        attention = layers.Flatten()(attention)

        model = layers.Dense(n_features)(attention)

        return Model(inputs, model)



import importlib.resources as pkg_resources
import json

def static_params():
    with pkg_resources.open_text('epicas.settings', 'hyperparams.json') as f:
        parameters = json.load(f)

    return parameters



"""
This module reads structured data (CSV, Excel, JSON, etc.) to Pandas dataframes
"""
import pandas as pd

from epicas import utils

class StructuredData:
    """
    # Arguments:
        file_input = str. Path or URL to the dataset.

        date = str or None.
            - Specify the name of column that will be used as date column

            - If None (default), data will be parsed as static data. Static data
            is the data that is not changed, presumed to be the same through out
            outbreaking season. e.g., population data, asthma mortality.

        location = str or None.
            - Specify the name of the location column in your data.
            - If None, dataset will be read as single location.

        incidence = str or None.
            - If specified, the name of the column that will be read as incident
            cases (new cases over an interval of days)

        prevalence = str or None.
            - If specified, the name of the column that will be read as prevalent
            cases (currently active cases at some specific time period)

        death = str or None.
            - If specified, the name of the column that will be read as deaths due
            to infection

        recovered = str or None.
            - If specified, the name of the column that will be read as recovered
            cases

        cumulative = bool. False by default.
            - If True, the values of incident, prevalent, recovered cases, and/or
            deaths will be read as "total recorded from start of outbreak to
            specified date."

        df = pandas.DataFrame

        **kwargs: other arguments can be passed into pandas.read_csv
            (see pandas documentation for all possible arguments).
    """

    def __init__(self, file_input = None, date = None, location = None,
                incidence = None, prevalence = None, death = None,
                recovered = None, cumulative = False, df = None,
                static_vars = None, time_series_vars=None, **kwargs):

        self.path = file_input
        self.date = date
        self.location = location
        self.incidence = incidence
        self.prevalence = prevalence
        self.death = death
        self.recovered = recovered
        self.cumulative = cumulative
        self.variables = {}
        self.variables['static'] = static_vars
        self.variables['time_series'] = time_series_vars

        headers = {
            self.date : 'date',
            self.location : 'location',
            self.incidence : 'incidence',
            self.prevalence : 'prevalence',
            self.recovered : 'recovered',
            self.death : 'death'
        }

        del headers[None]

        if isinstance(df, pd.DataFrame):
            self.df = df.rename(columns=headers)
            if 'date' in self.df.columns.values.tolist():
                self.df.date = pd.to_datetime(self.df.date, errors='coerce')

        else:
            self.df = pd.read_csv(self.path, **kwargs)
            self.df = self.df.rename(columns=headers)
            if 'date' in self.df.columns.values.tolist():
                self.df.date = pd.to_datetime(self.df.date, errors='coerce')

        if self.variables['static'] == None and self.date == None:
            self.variables['static'] = self.df.columns.values.tolist()

        elif self.variables['time_series'] == None:
            self.variables['time_series'] = self.df.columns.values.tolist()


    def __add__(self, next):
        """
        # Usage:
            df = StructuredData(...)
            df2 = StructuredData(...)
            df = df + df2
        """
        commons = list(self.df.columns.intersection(next.df.columns))

        joined = self.df.merge(next.df, on=commons)

        static_vars, time_series_vars = utils.variables_merge(
            self.variables['static'],
            self.variables['time_series'],
            next.variables['static'],
            next.variables['time_series']
        )

        return StructuredData(
            df = joined,
            static_vars = static_vars,
            time_series_vars = time_series_vars)


    def __str__(self):
        return f'StructuredData{list(self.df.columns)}\n\n' + \
                f'Variables: {self.variables}\n\n' + \
                f'{self.df}\n\n'


    def stats(self):
        """Examine StructuredData for NaN"""

        utils.count_na(self.df)


    def drop(self, drop_list):
        """Drop features from StructuredData

        # Argument:
            - drop_list: list. List of features to be dropped"""

        self.df = self.df.drop(columns=drop_list)


class UnstructuredData:
    """ To be developed """

    def __init__(**kwargs):
        pass
